{-# LANGUAGE OverloadedStrings #-}

-- | This module contains various pure functions for transforming the state of
-- 'IrcChannel' structures as a result of executing IRC commands.
module Pipes.IRC.Server.Channel
 ( newChannel
 , chanAddModeFlag
 , chanDelModeFlag
 , chanAddUser
 , chanDelUser
 , chanHasUser
 , chanSetTopic
 , chanHasTopic
 , chanHasModeFlag
 , chanSigil
 , chanUserSigil
 , chanSetPass
 , chanClearPass
 , chanHasPass
 , chanCheckPass
 , chanAddOper
 , chanDelOper
 , chanAddVoice
 , chanDelVoice
 , chanAddInvite
 , chanDelInvite
 , chanChangeNick
 , chanUserIsOper
 , chanUserHasVoice
 , chanUserIsInvited
 , chanUserMaySpeak
 , chanUserMayJoin
 , chanUserMaySetTopic
 )
where

import           Control.Lens
import           Data.ByteString.Char8  (ByteString)
import           Data.Maybe             (isJust)
import           Data.Set               (delete, empty, fromList, insert,
                                         member)

import           Pipes.IRC.Server.Types

-- | Create a new 'IrcChannel' record with the given nickname as the
-- only member.
newChannel :: NickKey    -- ^ the nickname of the channel creator
           -> IrcChannel -- ^ the resulting new 'IrcChannel' record
newChannel creator = IrcChannel { _chanTopic = Nothing
                                , _chanPass  = Nothing
                                , _chanModeFlags = empty
                                , _chanUsers = fromList [creator]
                                , _chanOpers = fromList [creator]
                                , _chanVoices = empty
                                , _chanInvites = empty
                                }

chanAddModeFlag, chanDelModeFlag :: IrcChanModeFlags
                                 -- ^ channel mode flag to change
                                 -> IrcChannel
                                 -- ^ channel on which to change mode
                                 -> IrcChannel
                                 -- ^ resulting channel with mode change
-- | Add a mode flag to the channel's set of mode flags. See
-- 'IrcChanModeFlags' for the available modes.
chanAddModeFlag cm = chanModeFlags %~ insert cm
-- | Remove the mode flag from the channel's set of mode flags. See
-- 'IrcChanModeFlags' for the available modes.
chanDelModeFlag cm = chanModeFlags %~ delete cm

chanAddUser, chanDelUser :: NickKey    -- ^ user to add to channel
                         -> IrcChannel -- ^ channel on which to add user
                         -> IrcChannel -- ^ channel with user added

-- | Add a user to the channel's list of users. This does not change
-- the set of set of channels on the user or perform any checking;
-- this just performs the low-level change to the channel.
chanAddUser uname = chanUsers %~ insert uname

-- | Delete a user from the channel's list of users. This does not
-- change the set of channels on the user or perform any checking;
-- this just performs the low-level change to the channel.
chanDelUser uname = (chanUsers %~ delete uname)
                  . (chanOpers %~ delete uname)
                  . (chanVoices %~ delete uname)

-- | Is the user owning the given nick on the channel? Returns 'True'
-- if the user is in fact on the channel.
chanHasUser :: NickKey    -- ^ user to check presence of on the channel
            -> IrcChannel -- ^ channel in which to look for the user
            -> Bool       -- ^ 'True' if the user is in the channel
chanHasUser uname ch = member uname (ch ^. chanUsers)

-- | Set the topic of discussion in the channel. This does not do any
-- permission or size checking; it just performs the low-level action.
chanSetTopic :: ByteString -- ^ topic of discussion for the channel
             -> IrcChannel -- ^ channel on which to set the topic
             -> IrcChannel -- ^ channel with the topic set
chanSetTopic top = chanTopic .~ Just top

-- | Does the channel have a topic of discussion set? Returns 'True'
-- if the topic is set.
chanHasTopic :: IrcChannel -- ^ channel on which to look for a topic
             -> Bool       -- ^ 'True' if the channel has a topic
chanHasTopic ch = isJust $ ch ^. chanTopic

-- | Does the channel have a specific mode flag set? Returns 'True'
-- if it has the specified flag set.
chanHasModeFlag :: IrcChanModeFlags -- ^ mode flag to check for
                -> IrcChannel       -- ^ channel on which to check
                -> Bool             -- ^ 'True' if the flag is set
chanHasModeFlag cm ch = member cm $ ch ^. chanModeFlags

-- | Find the mode sigil determined by what privacy modes it has
-- set, if any.
chanSigil :: IrcChannel -- ^ channel of which to find the sigil
          -> ByteString -- ^ the sigil associated with the channel
chanSigil ch | chanHasModeFlag Secret ch  = "@"
             | chanHasModeFlag Private ch = "*"
             | otherwise                  = "="

-- | Find the mode sigil determined by a user's status on a channel,
-- e.g. "@" for a channel op or "+" for a user with voice status.
chanUserSigil :: NickKey    -- ^ nickname of user of which to find sigil
              -> IrcChannel -- ^ channel the user may have status in
              -> ByteString -- ^ the sigil associated with the user's status
chanUserSigil uname ch | member uname $ ch ^. chanOpers  = "@"
                       | member uname $ ch ^. chanVoices = "+"
                       | otherwise                       = ""

-- | Set a password key that must be given with a JOIN command in
-- order to join the channel. This does no permission or error
-- checking, it just performs the low-level operation.
chanSetPass :: PassKey    -- ^ password key to set for a channel
            -> IrcChannel -- ^ channel to set the password key on
            -> IrcChannel -- ^ channel with the password set
chanSetPass pwd = chanPass .~ Just pwd

-- | Clear password key for the channel. This does not do permission or
-- error checking, it just performs the low-level operation.
chanClearPass :: IrcChannel -- ^ channel to clear the password key
              -> IrcChannel -- ^ channel with the password cleared
chanClearPass = chanPass .~ Nothing

-- | Does the channel have a password key set? Returns 'True' if the
-- channel does have a password key set.
chanHasPass :: IrcChannel -- ^ channel on which to look for password key
            -> Bool       -- ^ 'True' if there is a password key set
chanHasPass ch = isJust $ ch ^. chanPass

-- | Determine whether the given password key should give access to
-- the channel. If it matches the set password or there is no
-- password, the response will be 'True'.
chanCheckPass :: ByteString -- ^ password to try against the channel
              -> IrcChannel -- ^ channel against which to try the password
              -> Bool       -- ^ 'True' if access should be granted
chanCheckPass pwd ch = case ch ^. chanPass of
  Just chPass -> pwd == chPass
  Nothing     -> True

chanAddOper, chanDelOper :: NickKey    -- ^ nickname of user
                         -> IrcChannel -- ^ channel to change status on
                         -> IrcChannel -- ^ channel with status changed

-- | Add the given nickname as an operator on the channel. This does
-- not perform any permission or error checks, it just performs the
-- low-level operation.
chanAddOper uname = chanOpers %~ insert uname

-- | Remove the given nickname as an operator on the channel. This
-- does not perform any permission or error checks, it just performs
-- the low-level operation.
chanDelOper uname = chanOpers %~ delete uname

chanAddVoice, chanDelVoice :: NickKey    -- ^ nickname of user
                           -> IrcChannel -- ^ channel to change voice on
                           -> IrcChannel -- ^ changed channel

-- | Give voice status to the given nickname on the channel. This
-- does not perform any permission or error checks, it just performs
-- the low-level operation.
chanAddVoice uname = chanVoices %~ insert uname

-- | Remove voice status from the given nickname on the channel. This
-- does not perform any permission or error checks, it just performs
-- the low-level operation.
chanDelVoice uname = chanVoices %~ delete uname

chanAddInvite, chanDelInvite :: NickKey    -- ^ nickname of user
                             -> IrcChannel -- ^ channel to change invitation
                             -> IrcChannel -- ^ channel with change applied

-- | Add a record of invitation for the nickname to join the channel.
-- This will allow the user to join when the channel is invite-only.
-- This does not perform any permission or error checks, it just does
-- the operation.
chanAddInvite uname = chanInvites %~ insert uname

-- | Remove record of the invitation for the nickname to join the
-- channel. This does not perform permission or error checks, it just
-- does the low-level operation.
chanDelInvite uname = chanInvites %~ delete uname

-- | Change the nick of a user on the channel. This will replace the
-- old name with the news name in the member list, invitation list,
-- and the mode flags for privileges. It will only affect the
-- channel, not the user or other server records.
chanChangeNick :: NickKey    -- ^ old nickname
               -> NickKey    -- ^ new nickname
               -> IrcChannel -- ^ channel to modify
               -> IrcChannel -- ^ channel with nickname changed
chanChangeNick old new ch = (chOps . chVoice . chInvite . chUsers) ch
  where
   chOps
     | chanUserIsOper old ch = chanDelOper old . chanAddOper new
     | otherwise             = id
   chVoice
     | chanUserHasVoice old ch = chanDelVoice old . chanAddVoice new
     | otherwise                = id
   chInvite
     | chanUserIsInvited old ch = chanDelInvite old . chanAddInvite new
     | otherwise                = id
   chUsers
     | chanHasUser old ch = chanDelUser old . chanAddUser new
     | otherwise          = id

-- | Is the nickname an operator of the channel? Returns 'True' if
-- the user is a channel op. This does not consider any other
-- privileges.
chanUserIsOper :: NickKey    -- ^ nickname to check
               -> IrcChannel -- ^ channel to check
               -> Bool       -- ^ 'True' if nickname has ops on channel
chanUserIsOper uname ch = member uname $ ch ^. chanOpers

-- | Does the nickname have voice on the channel? Returns 'True' if
-- the user has the voice mode flag. This does not consider any other
-- privileges or conditions.
chanUserHasVoice :: NickKey    -- ^ nickname to check
                 -> IrcChannel -- ^ channel to check
                 -> Bool       -- ^ 'True' if nickname has voice on channel
chanUserHasVoice uname ch = member uname $ ch ^. chanVoices

-- | Has the nickname been invited to the channel? Returns 'True' if
-- the user has been invited. This does not consider any other
-- privileges or conditions.
chanUserIsInvited :: NickKey    -- ^ nickname to check
                  -> IrcChannel -- ^ channel to check
                  -> Bool       -- ^ 'True' if nickname has been invited
chanUserIsInvited uname ch = member uname $ ch ^. chanInvites

-- | Is the nickname allowed to speak on the channel? Returns 'True'
-- if the relevant permission checks all pass. The checks consider
-- channel membership, the NoOutsideMsgs flag, the Moderated flag and
-- Voice status, and channel Operator status.
chanUserMaySpeak :: NickKey    -- ^ nickname to check
                 -> IrcChannel -- ^ channel to check
                 -> Bool       -- ^ 'True' if nickname may speak on channel
chanUserMaySpeak uname ch
  | (chanHasUser uname ch || not (chanHasModeFlag NoOutsideMsgs ch))
     && not (chanHasModeFlag Moderated ch) = True
  | chanUserIsOper uname ch                = True
  | chanUserHasVoice uname ch              = True
  | otherwise                              = False

-- | Is the nickname allowed to join the channel? Returns 'True' if
-- the relevant permission checks all pass. The checks consider the
-- InviteOnly flag and whether an invitation has been issued.
chanUserMayJoin :: NickKey    -- ^ nickname to check
                -> IrcChannel -- ^ channel to check
                -> Bool       -- ^ 'True' if nickname may join the channel
chanUserMayJoin uname ch
  | not $ chanHasModeFlag InviteOnly ch = True
  | chanUserIsInvited uname ch          = True
  | otherwise                           = False

-- | Is the nickname allowed to set the topic for the channel?
-- Returns 'True' if the relevant permission checks all pass.
chanUserMaySetTopic :: NickKey    -- ^ nickname to check
                    -> IrcChannel -- ^ channel to check
                    -> Bool       -- ^ 'True' if nickname may set the topic
chanUserMaySetTopic uname ch
  | not (chanHasModeFlag TopicOperOnly ch) &&
    chanHasUser uname ch    = True
  | chanUserIsOper uname ch = True
  | otherwise               = False
