-- | This module contains various utility functions that don't fit elsewhere.
module Pipes.IRC.Server.Util
  ( parseParamList
  , zipParams
  , adjustAtKeys
  , alterAtKeys
  )
where

import           Data.ByteString.Char8 (ByteString, null, split)
import           Data.Map              (Map, adjust, alter)
import           Prelude               hiding (null)

-- | Split a 'ByteString' containing an IRC command parameter list into a
-- list of 'ByteString' parameters.
parseParamList :: ByteString   -- ^ comma-delimited string of parameters
               -> [ByteString] -- ^ list of parameter strings
parseParamList ps = filter (not . null) $ split ',' ps

-- | Pair a list of parameters (such as channel names) with a corresponding
-- second list of parameters (such as channel passwords) they should be
-- paired with. The second list may be empty or up to the same length as the
-- first list; differences will be made up with 'Nothing' values.
zipParams :: [ByteString] -- ^ first list of parameter strings
          -> [ByteString] -- ^ list of parameters to pair them with
          -> [(ByteString, Maybe ByteString)]
                          -- ^ list of parameter pairs
zipParams chans chkeys = zip chans (map Just chkeys ++ repeat Nothing)

-- | A "Data.Map" helper that applies an 'adjust' function at each of the
-- given keys. For example:
--
-- >>> adjustAtKeys (++ "x") [1,3] (fromList [(1, "a"), (2, "b"), (3, "c")])
-- fromList [(1, "ax"), (2, "b"), (3, "cx")]
--
adjustAtKeys :: Ord k
             => (v -> v) -- ^ function to apply with 'adjust'
             -> [k]      -- ^ keys to apply the adjustment at
             -> Map k v  -- ^ map to apply the adjustments to
             -> Map k v  -- ^ new map resulting from the adjustments
adjustAtKeys f keys dmap = foldr (adjust f) dmap keys

-- | A "Data.Map" helper that applies an 'alter' function at each of the
-- given keys. This is similar to 'adjustAtKeys' except it may create or
-- remove key/value pairs instead of just changing existing ones.
alterAtKeys :: Ord k
            => (Maybe v -> Maybe v) -- ^ function to apply with 'alter'
            -> [k]      -- ^ keys to apply the alteration at
            -> Map k v  -- ^ map to apply the alterations to
            -> Map k v  -- ^ new map resulting from the alterations
alterAtKeys f keys dmap = foldr (alter f) dmap keys
